#!/usr/bin/python

import os, re, argparse, json
import xml.etree.ElementTree as et
from collections import OrderedDict

# ############################################################################ #
# Global values
# ############################################################################ #
verbose = True
layer_visibility = ""

# ############################################################################ #
# Utility methods
# ############################################################################ #
def log(msg):
    """ Prints the provided message to standard output unless the script is run
        in silent mode.
    """
    global verbose
    if verbose == True:
        print msg

def join_path(wdir, path):
    """ Prepends provided path with wdir unless path begins with a slash or a 
        capital letter followed by a colon. In this case it is treated as an
        absolute path and its original value is returned.
    """
    if (path is None):
        return ""
    if path.startswith(r'/') or path.startswith(r'[A-Z]:'):
        return path
    else:
        return os.path.join(wdir, path)

# ############################################################################ #
# Extracting keys
# ############################################################################ #
def find_keys(file_path):
    """ Extracts service keys from the file at the given path, as long as the
        file is either Info.plist or AndroidManifest.xml.
    """
    ios_re = re.compile(r'.*Info\.plist$')
    android_re = re.compile(r'.*AndroidManifest\.xml$')

    if (file_path is None):
        raise Exception("No service key file path provided.")

    if ios_re.match(file_path):
        return find_keys_ios(file_path)
    elif android_re.match(file_path):
        return find_keys_android(file_path)
    else:
        raise Exception("The service key file path provided points to neither "
                        "AndroidManifest.xml nor Info.plist")

def find_keys_android(file_path):
    """ Extracts service keys from AndroidManifest.xml.
    """
    manifest = et.parse(file_path)
    root = manifest.getroot()
    mdtags = root.find("application").findall("meta-data")
    maps_key = None
    traffic_key = None

    for tag in mdtags:
        if tag.get("{http://schemas.android.com/apk/res/android}name") == "OnlineMaps.Key":
            maps_key = tag.get("{http://schemas.android.com/apk/res/android}value")
        if tag.get("{http://schemas.android.com/apk/res/android}name") == "OnlineTraffic.Key":
            traffic_key = tag.get("{http://schemas.android.com/apk/res/android}value")

    if (maps_key == None or traffic_key == None):
        raise Exception("Missing service keys.")
        
    return maps_key, traffic_key

def find_keys_ios(file_path):
    """ Extracts service keys from Info.plist
    """
    maps_re = re.compile(
        r'<key>OnlineMap.Key</key>\W*<string>([a-zA-Z0-9]*)</string>')
    traffic_re = re.compile(
        r'<key>OnlineTraffic.Key</key>\W*<string>([a-zA-Z0-9]*)</string>')

    return find_keys_by_regex(file_path, maps_re, traffic_re)

def find_keys_by_regex(file_path, maps_re, traffic_re):
    """ Searches file at provided file_path for strings matching maps_re
        and traffic_re.
    """
    with open(file_path, 'r') as file:
        data = file.read()

        maps_key_match = maps_re.search(data)
        traffic_key_match = traffic_re.search(data)

        if maps_key_match is not None and traffic_key_match is not None:
            maps_key = maps_key_match.group(1)
            traffic_key = traffic_key_match.group(1)
            if maps_key is not None and traffic_key is not None:
                return maps_key, traffic_key

    raise Exception("Missing service keys.")

# ############################################################################ #
# Conversion to maputnik format
# ############################################################################ #
def convert_for_maputnik(
        input_path, output_path, maps_key, traffic_key, glyph_url, sprite_url):
    """ Replaces tomtom-specific values in the file at input_path with provided
        values so that it can be loaded and edited in maputnik.
    """
    output = ""
    with open(input_path, 'r') as input_file:
        output = input_file.read()

    # 1. Replace online maps key tags with actual key values
    maps_key_tag_re = re.compile(r'\[OnlineMaps\]')
    output = maps_key_tag_re.sub(maps_key, output)
    
    # 2. Replace online traffic key tags with actual key values
    traffic_key_tag_re = re.compile(r'\[OnlineTraffic\]')
    output = traffic_key_tag_re.sub(traffic_key, output)

    # 3. Remove language and view tags
    remove_kay_tag_re = re.compile(r'(&language=\[Language\]|&view=\[GeoView\])')
    output = remove_kay_tag_re.sub(r'', output)

    # 4. Replace flow tag
    flow_tag_re = re.compile(r'\[FlowStyle\]')
    output = flow_tag_re.sub(r'relative', output)

    # 5. Replace glyph and sprite links
    glyph_line_re = re.compile(r'"glyphs": "[^"]*"')
    output = glyph_line_re.sub(
        r'"glyphs": "' + glyph_url + r'"',
        output)

    sprite_line_re = re.compile(r'"sprite": "[^"]*"')
    output = sprite_line_re.sub(
        r'"sprite": "' + sprite_url + r'"',
        output)
    
    # 6. Change visibility
    output = change_visibility(output)

    with open(output_path, 'w') as output_file:
        output_file.write(output)

def change_visibility(output):
    all_visibility_lines_re = re.compile(r'"visibility": "[a-zA-Z]*"')
    if layer_visibility == "all":
        output = all_visibility_lines_re.sub(
            r'"visibility": "visible"',
            output)
    elif layer_visibility == "none":
        output = all_visibility_lines_re.sub(
            r'"visibility": "none"',
            output)
    elif layer_visibility == "predefined":
        root = json.loads(output, object_pairs_hook=OrderedDict)

        layers = root[r'layers']

        for layer in layers:
            if r'source' in layer and r'layout' in layer:
                if r'visibility' in layer[r'layout']:
                    if layer[r'source'] == r'vectorTiles':
                        layer[r'layout'][r'visibility'] = r'visible'
                    else:
                        layer[r'layout'][r'visibility'] = r'none'

        output = json.dumps(root, indent=2)
        
    return output

# ############################################################################ #
# Conversion to tomtom format
# ############################################################################ #
def convert_for_tomtom(base_path, edited_path, out_path):
    """ Replaces changes in tomtom-specific elements of the style edited in
        maputnik with their original values.
    """
    base = ""
    edited = ""

    # Open and read the base file
    with open(base_path, 'r') as base_file:
        base = base_file.read()

    # Open and read the edited file
    with open(edited_path, 'r') as edited_file:
        edited = edited_file.read()

    edited = revert_tile_urls(base, edited)
    edited = revert_sprites_url(base, edited)
    edited = revert_glyphs_url(base, edited)

    # Save the output
    with open(out_path, 'w') as out_file:
        out_file.write(edited)
        return
    return

def revert_tile_urls(base, edited):
    regex = re.compile(r'"https://.*api.tomtom.com/[a-zA-Z0-9/]*(?={|\[)')
    return revert_by_regex(base, edited, regex)

def revert_sprites_url(base, edited):
    regex = re.compile(r'"glyphs":')
    return revert_by_regex(base, edited, regex)

def revert_glyphs_url(base, edited):
    regex = re.compile(r'"sprite":')
    return revert_by_regex(base, edited, regex)

def revert_by_regex(base, edited, regex):
    matching_lines = regex.findall(base)

    for line in matching_lines:
        pattern = line + r'.*"'
        line_re = re.compile(pattern)
        match = line_re.search(base)
        replacement = match.group()
        edited = line_re.sub(replacement, edited)

    return edited

# ############################################################################ #
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-w",
        "--wdir",
        default="./",
        help="Working directory. Root of relative paths.")
    parser.add_argument(
        "-S",
        "--silent",
        action="store_true",
        help="Silent mode. No messages will be printed.")
    parser.add_argument(
        "-b",
        "--base",
        required=True,
        help="Path to the base for the new style.")
    parser.add_argument(
        "-o",
        "--output",
        required=True,
        help="Path to the output file.")

    subparsers = parser.add_subparsers(dest="command")

    # Parser for "maputnik" command
    maputnik_parser = subparsers.add_parser(
        "maputnik",
        help="Convert a base style so that it can be edited in maputnik."
    )

    maputnik_parser.add_argument(
        "-k",
        "--keys",
        required=True,
        help="Path to Info.plist or AndroiManifest.xml.")
    maputnik_parser.add_argument(
        "-s",
        "--sprites",
        required=True,
        help="Online location of sprites.")
    maputnik_parser.add_argument(
        "-g",
        "--glyphs",
        required=True,
        help="Online location of glyphs.")
    maputnik_parser.add_argument(
        "-v",
        "--visibility",
        choices=["all", "none", "intact", "predefined"],
        default="predefined",
        help="Change visibility of layers. 'intact' leaves the base style's"
             " settings intact, 'predefined' makes vector layers visible"
             " and all other layers invisible. Default = 'predefined'"
    )

    # Parser for "tomtom" command
    tomtom_parser = subparsers.add_parser(
        "tomtom",
        help="Convert a style edited in maputnik so that it works with your "
             "application using TomTom SDK."
    )

    tomtom_parser.add_argument(
        "-e",
        "--edited",
        required=True,
        help="Path to the style edited in maputnik."
    )

    args = parser.parse_args()

    # 1. Set verbosity
    global verbose
    verbose = not args.silent

    # 2. Create paths to base and output files
    base_path = join_path(wdir=args.wdir, path=args.base)
    out_path = join_path(wdir=args.wdir, path=args.output)

    try:
        if args.command == "maputnik":
            global layer_visibility
            layer_visibility = args.visibility
            # 3. Extract keys
            maps_key, traffic_key = find_keys(
                    join_path(wdir=args.wdir, path=args.keys))

            print maps_key
            print traffic_key

            # 4. Convert the file to maputnik's format
            glyph_url = args.glyphs
            sprite_url = args.sprites
            convert_for_maputnik(
                    base_path,
                    out_path,
                    maps_key,
                    traffic_key,
                    glyph_url,
                    sprite_url)
            return
        elif args.command == "tomtom":
            # 3. Convert the edited file to tomtom's format
            edited_path = join_path(wdir=args.wdir, path=args.edited)
            convert_for_tomtom(
                    base_path,
                    edited_path,
                    out_path
            )
            return
        else:
            log("Invalid command.")
    except Exception as e:
        log("Exception: {0}".format(e))

if __name__ == "__main__":
    main()
